unit ifpnm;

interface

uses ImgFilter;

type
  TPnmFilter = class(TImgFilter)
    constructor Create; override;
    function Test(Buf: PChar; BufSize: Cardinal): Boolean; override;
    procedure GetSize(Buf: PChar; BufSize: Cardinal;
      var Width: Cardinal; var Height: Cardinal;
      var Components: Cardinal); override;
    procedure Decode(InBuf: PChar; InBufSize: Cardinal; OutBuf: PChar;
      OutBufSize: Cardinal; var ImgInfo: TImgInfo); override;
  end;

implementation

uses NLib;

type
  TPnmInfo = record
    Format: Char;
    Width: Cardinal;
    Height: Cardinal;
    MaxVal: Cardinal;
    Comment: string;
  end;

const
  ChType: array[Char] of Byte = (
    0,0,0,0,0,0,0,0,0,1,1,0,0,1,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
  );
  Space = 1;
  Digit = 2;

  MinPnmSize = 8; // Minimum PNM is 8bytes: "P1 1 1 1"

function GetVal(Buf: PChar; EndP: PChar; var Value: Cardinal): PChar;
begin
  while (Buf < EndP) and (ChType[Buf^] = Space) do
    Inc(Buf);

  Value := 0;
  while (Buf < EndP) and (ChType[Buf^] = Digit) do
  begin
    Value := Value * 10 + Ord(Buf^) - Ord('0');
    Inc(Buf);
  end;

  Result := Buf;
end;

function GetValBw(Buf: PChar; EndP: PChar; var Value: Cardinal): PChar;
begin
  while (Buf < EndP) and (ChType[Buf^] = Space) do
    Inc(Buf);

  if Buf < EndP then
  begin
    Value := Ord(Buf^) - Ord('0');
    Inc(Buf);
  end else
    Value := 0;

  Result := Buf;
end;

// return pointer of bit data. if failed, return nil
function GetHeader(Buf: PChar; BufSize: Cardinal; var Info: TPnmInfo): PChar;
var
  p, ep: PChar;
  i, tmp, n: Cardinal;
begin
  Result := nil;

  if BufSize < MinPnmSize then
    Exit;

  Info.Width := 0;
  Info.Height := 0;
  Info.Comment := '';
  Info.Format := Buf[1];
  if not (Info.Format in ['1'..'6']) then
    Exit;

  if Info.Format in ['1', '4'] then
  begin
    n := 2;
    Info.MaxVal := 1;
  end else
  begin
    n := 3;
    Info.MaxVal := 0;
  end;

  ep := @Buf[BufSize];

  // skip magic number
  while (Buf < ep) and (ChType[Buf^] <> Space) do
    Inc(Buf);

  i := 0;
  while (Buf < ep) and (i < n) do
  begin
    if ChType[Buf^] = Space then
      Inc(Buf)
    else if Buf^ = '#' then
    begin
      p := Buf;
      while Buf^ <> #10 do
        Inc(Buf);
      Info.Comment := Info.Comment + Copy(p, 2, Buf - p);
    end else
    begin
      Buf := GetVal(Buf, ep, tmp);
      case i of
        0: Info.Width := tmp;
        1: Info.Height := tmp;
        2: Info.MaxVal := tmp;
      end;
      Inc(i);
    end;
  end;

  if (Info.Width <> 0) and (Info.Height <> 0) and (Info.MaxVal <> 0) then
    Result := Buf+1;
end;

// PBM ascii
function DecodeP1(Buf: PChar; EndBuf: PChar; const Info: TPnmInfo; Line: PChar): PChar;
var
  x, tmp: Cardinal;
begin
  Result := Buf;
  for x := 0 to Info.Width-1 do
  begin
    Result := GetValBw(Result, EndBuf, tmp);
    Line^ := Char(tmp + 255);
    Inc(Line);
    if Result >= EndBuf then
      Exit;
  end;
end;

// PGM ascii
function DecodeP2(Buf: PChar; EndBuf: PChar; const Info: TPnmInfo; Line: PChar): PChar;
var
  x, tmp, max: Cardinal;
begin
  Result := Buf;
  max := Info.MaxVal;
  if max = 255 then
    for x := 0 to Info.Width-1 do
    begin
      if Result >= EndBuf then
        Exit;
      Result := GetVal(Result, EndBuf, tmp);
      Line^ := Char(tmp);
      Inc(Line);
    end
  else
    for x := 0 to Info.Width-1 do
    begin
      if Result >= EndBuf then
        Exit;
      Result := GetVal(Result, EndBuf, tmp);
      Line^ := Char((tmp * 255 + max shr 1) div max);
      Inc(Line);
    end
end;

// PPM ascii
function DecodeP3(Buf: PChar; EndBuf: PChar; const Info: TPnmInfo; Line: PChar): PChar;
var
  x, r, g, b, max: Cardinal;
begin
  Result := Buf;
  max := Info.MaxVal;
  if max = 255 then
    for x := 0 to Info.Width-1 do
    begin
      if Result >= EndBuf then
        Exit;
      Result := GetVal(Result, EndBuf, r);
      Result := GetVal(Result, EndBuf, g);
      Result := GetVal(Result, EndBuf, b);
      Line[0] := Char(b);
      Line[1] := Char(g);
      Line[2] := Char(r);
      Inc(Line, 3);
    end
  else
    for x := 0 to Info.Width-1 do
    begin
      if Result >= EndBuf then
        Exit;
      Result := GetVal(Result, EndBuf, r);
      Result := GetVal(Result, EndBuf, g);
      Result := GetVal(Result, EndBuf, b);
      Line[0] := Char((b * 255 + max shr 1) div max);
      Line[1] := Char((g * 255 + max shr 1) div max);
      Line[2] := Char((r * 255 + max shr 1) div max);
      Inc(Line, 3);
    end;
end;

// PBM rawbits
function DecodeP4(Buf: PChar; EndBuf: PChar; const Info: TPnmInfo; Line: PChar): PChar;
var
  x, m: Cardinal;
begin
  Result := Buf;
  m := 0;
  for x := 0 to Info.Width-1 do
  begin
    if Result >= EndBuf then
      Exit;
    m := 7 - x and 7; // place of bit
    Line^ := Char((Byte(Result^) shr m and 1 + 255));
    Inc(Line);
    if m = 0 then
      Inc(Result);
  end;
  if m > 0 then // 8bit align
    Inc(Result);
end;

// PGM rawbits
function DecodeP5(Buf: PChar; EndBuf: PChar; const Info: TPnmInfo; Line: PChar): PChar;
var
  x, max: Cardinal;
begin
  Result := Buf;
  max := Info.MaxVal;
  if max = 255 then
    for x := 0 to Info.Width-1 do
    begin
      if Result >= EndBuf then
        Exit;
      Line^ := Result^;
      Inc(Line);
      Inc(Result);
    end
  else if max < 255 then
    for x := 0 to Info.Width-1 do
    begin
      if Result >= EndBuf then
        Exit;
      Line^ := Char((Byte(Result^) * 255 + max shr 1) div max);
      Inc(Line);
      Inc(Result);
    end
  else
    for x := 0 to Info.Width-1 do
    begin
      if Result >= EndBuf then
        Exit;
      Line^ := Char(((Byte(Result[0]) shl 8 or Byte(Result[1])) * 255 + max shr 1) div max);
      Inc(Line);
      Inc(Result, 2);
    end;
end;

// PPM rawbits
function DecodeP6(Buf: PChar; EndBuf: PChar; const Info: TPnmInfo; Line: PChar): PChar;
var
  x, max: Cardinal;
begin
  Result := Buf;
  max := Info.MaxVal;
  if max = 255 then
    for x := 0 to Info.Width-1 do
    begin
      if Result > EndBuf-3 then
        Exit;
      Line^ := Result[2];
      Inc(Line);
      Line^ := Result[1];
      Inc(Line);
      Line^ := Result[0];
      Inc(Line);
      Inc(Result, 3);
    end
  else if max < 255 then
    for x := 0 to Info.Width-1 do
    begin
      if Result > EndBuf-3 then
        Exit;
      Line[0] := Char((Byte(Result[2]) * 255 + max shr 1) div max);
      Line[1] := Char((Byte(Result[1]) * 255 + max shr 1) div max);
      Line[2] := Char((Byte(Result[0]) * 255 + max shr 1) div max);
      Inc(Line, 3);
      Inc(Result, 3);
    end
  else
    for x := 0 to Info.Width-1 do
    begin
      if Result > EndBuf-3 then
        Exit;
      Line[0] := Char(((Byte(Result[4]) shl 8 or Byte(Result[5])) * 255 + max shr 1) div max);
      Line[1] := Char(((Byte(Result[2]) shl 8 or Byte(Result[3])) * 255 + max shr 1) div max);
      Line[2] := Char(((Byte(Result[0]) shl 8 or Byte(Result[1])) * 255 + max shr 1) div max);
      Inc(Line, 3);
      Inc(Result, 6);
    end;
end;


{ TPnmFilter }

constructor TPnmFilter.Create;
begin
  FName := '(Internal PNM to DIB filter)';
end;

function TPnmFilter.Test(Buf: PChar; BufSize: Cardinal): Boolean;
begin
  Result := (BufSize >= MinPnmSize) and
    (Buf[0] = 'P') and
    (Buf[1] in ['1'..'6']) and
    (Buf[2] in [' ', '#', #9, #13, #10]);
end;

procedure TPnmFilter.GetSize(Buf: PChar; BufSize: Cardinal;
  var Width: Cardinal; var Height: Cardinal;
  var Components: Cardinal);
var
  PnmInfo: TPnmInfo;
begin
  if GetHeader(Buf, BufSize, PnmInfo) = nil then
    raise EImgFilterError.Create('Invalid PNM header');

  Width := PnmInfo.Width;
  Height := PnmInfo.Height;
  if PnmInfo.Format in ['3', '6'] then // PPM
    Components := 3
  else
    Components := 1;
end;

procedure TPnmFilter.Decode(InBuf: PChar; InBufSize: Cardinal; OutBuf: PChar;
  OutBufSize: Cardinal; var ImgInfo: TImgInfo);
type
  TDecodeProc = function (Buf: PChar; EndBuf: PChar; const Info: TPnmInfo;
    Line: PChar): PChar;
const
  FormatNames: array[0..6] of PChar = (
    '',
    'PBM image text',
    'PGM image text',
    'PPM image text',
    'PBM "rawbits" image data',
    'PGM "rawbits" image data',
    'PPM "rawbits" image data'
  );
  DecodeProcs: array[0..6] of TDecodeProc = (
    nil, DecodeP1, DecodeP2, DecodeP3, DecodeP4, DecodeP5, DecodeP6
  );
var
  PnmInfo: TPnmInfo;
  buf, line, endp: PChar;
  y, pitch, fmt: Cardinal;
begin
  buf := GetHeader(InBuf, InBufSize, PnmInfo);
  if buf = nil then
    raise EImgFilterError.Create('Invalid PNM header');

  if PnmInfo.Format in ['3', '6'] then // PPM
    pitch := (PnmInfo.Width * 3 + 3) and not 3
  else
    pitch := (PnmInfo.Width + 3) and not 3;
  line := @OutBuf[(PnmInfo.Height-1) * pitch]; // last line
  endp := @InBuf[InBufSize];
  fmt := Ord(PnmInfo.Format) - Ord('0');

  for y := 0 to PnmInfo.Height - 1 do
  begin
    buf := DecodeProcs[fmt](buf, endp, PnmInfo, line);
    Dec(line, pitch);
  end;

  ImgInfo.FormatName := FormatNames[fmt];
  if PnmInfo.Format in ['2', '3', '5', '6'] then
    ImgInfo.FormatName := NLib.Fmt('{0}, Maxval={1}',
      [ImgInfo.FormatName, PnmInfo.MaxVal]);
  ImgInfo.Comment := PnmInfo.Comment;
  case PnmInfo.Format of
    '1', '4': ImgInfo.Bpp := 1;
    '2', '5': ImgInfo.Bpp := 8;
    '3', '6': ImgInfo.Bpp := 24;
  end;
  if PnmInfo.MaxVal > 255 then
    ImgInfo.Bpp := ImgInfo.Bpp * 2;
end;

end.
